"""
Package for downloading data from the Arthur S3 bucket

.. warning::
    This package is experimental and is under active development. The API for this package (its specific methods and
    workflows) are subject to change as the design and scope of this package evolve.
"""
from .arthur_example import ArthurExample
